local casef = require "casefunc"
local ast = require "ast"
local Seq = require "seq"
local Set = require "set"
local ptype = require "printtype"
local tequal = require "eqtype"
local tparams = require "tparams"
local TypeEnv = require "tenv"
local replace = require "reptype"
local unify = require "unify"
local concrete = require "concrete"

local checkprim = {}

local function generalize(t, tenv)
    if t.tag ~= "tgen" then
        local params = Set:new():addAll(tparams(t)):
                                 removeAll(tenv:params())
        if #params > 0 then
            return ast.tgen{ lin = t.lin, col = t.col,
                             params = params, type = t }
        end
    end
    return t
end

local function especialize(tgen, tother, lin, col)
    if tgen.tag ~= "tgen" then
        return tgen
    end
    local r = {}
    for _, param in ipairs(tgen.params) do
        r[param.name] = ast.tvar{}
    end
    local t = replace(tgen.type, r)
    unify(t, tother, lin, col)
    return concrete(t)
end

local function especialize_func(tgen, targs, lin, col)
    if tgen.tag ~= "tgen" then
        return tgen
    end
    if not tgen.type.tag == "tfunc" then
        error("esperado função, mas dado " .. ptype(tgen.type) ..
              " na linha " .. lin .. " coluna " .. col)
    end
    local r = {}
    for _, param in ipairs(tgen.params) do
        r[param] = ast.tvar{}
    end
    local tfunc = replace(tgen.type, r)
    local tother = ast.tfunc{
            params = ast.types{ seq = Seq:new(targs) },
            ret = tfunc.ret
        }
    unify(tfunc, tother, lin, col)
    return concrete(tfunc)
end

function checkprim.seq(targs, lin, col)
    if #targs == 0 then
        return generalize(ast.tseq{ lin = lin, col = col,
                                    elem = ast.tparam{ lin = lin,
                                                       col = col,
                                                       name = "a" } },
                          TypeEnv:new())
    end
    local tprimeiro = targs[1]
    for i = 2, #targs do
        if not tequal(tprimeiro, especialize(targs[i], tprimeiro, lin, col)) then
            error("tipo do elemento " .. i .. " (" .. targs[i] ..
                  ") incompatível " .. " com " .. tprimeiro ..
                  " na linha " .. lin .. " coluna " .. col)
        end
    end
    return ast.tseq{ lin = lin, col = col, elem = tprimeiro }
end

local checktype = casef("checktype")

function checktype.stats(node, tenv)
    local t
    for _, cmd in ipairs(node.seq) do
        t, tenv = checktype(cmd, tenv)
        if not tequal(t, "tunit") then
            return t, tenv
        end
    end
    return t, tenv
end

checktype["while"] = function (node, tenv)
    local tcond = checktype(node.cond, tenv)
    if not tequal(tcond, "tbool") then
        error("condição do while não é booleana na linha " ..
            node.lin .. " coluna " .. node.col)
    end
    checktype(node.body, tenv)
    return ast.tunit{ lin = node.lin, col = node.col }, tenv
end

checktype["if"] = function (node, tenv)
    local tcond = checktype(node.cond, tenv)
    if not tequal(tcond, "tbool") then
        error("condição do if não é booleana na linha " ..
            node.lin .. " coluna " .. node.col)
    end
    local tthen, _ = checktype(node["then"], tenv)
    if node["else"] then
        local telse, _ = checktype(node["else"], tenv)
        if not tequal(tthen, "tunit") and
           not tequal(telse, "tunit") then
            return tthen, tenv
        end 
    end
    return ast.tunit{ lin = node.lin, col = node.col }, tenv
end

function checktype.def(node, tenv)
    return ast.tunit{ lin = node.lin, col = node.col },
           tenv:extend(node.var, 
               generalize(checktype(node.exp, tenv), tenv))
end

function checktype.func(node, tenv)
    local tenvb = tenv
    local ptypes = {}
    for _, prm in ipairs(node.params.seq) do
        tenvb = tenvb:extend(prm.name, prm.type)
        ptypes[#ptypes+1] = prm.type
    end
    local tfunc = ast.tfunc{ lin = node.lin,
                   col = node.col, ret = node.tret,
                   params = ast.types{ seq = Seq:new(ptypes) } }
    local tprms = tparams(tfunc)
    tenvb = tenvb:extend(node.name, tfunc)
    tenvb = tenvb:extend("$ret", node.tret)
    local tret = checktype(node.body, tenvb)
    if not tequal(tret, node.tret) then
        error("nem todos os caminhos estão retornando " ..
            ptype(node.tret) .. " na função na linha " .. node.lin ..
            " coluna " .. node.col)
    end
    return ast.tunit{ lin = node.lin, col = node.col },
           tenv:extend(node.name, generalize(tfunc, tenv))
end

function checktype.lmb(node, tenv)
    local tenvb
    local ptypes = {}
    for _, prm in ipairs(node.params.seq) do
        tenvb = tenv:extend(prm.name, prm.type)
        ptypes[#ptypes+1] = prm.type
    end
    tenvb = tenvb:extend("$ret", node.tret)
    local tret = checktype(node.body, tenvb)
    if not tequal(tret, node.tret) then
        error("nem todos os caminhos estão retornando " ..
            node.tret .. " na função na linha " .. node.lin ..
            " coluna " .. node.col)
    end
    return ast.tfunc{ lin = node.lin,
                      col = node.col, ret = node.tret,
                      params = ast.types{ seq = Seq:new(ptypes) } }
end

checktype["return"] = function (node, tenv)
    local t = checktype(node.exp, tenv)
    local tret = tenv:lookup("$ret", node.lin, node.col)
    if not tequal(tret, t) then
        error("tipo da expressão do return " .. ptype(t) .. 
            " não bate com tipo de retorno " .. ptype(tret) ..
            " na linha " .. node.lin .. " coluna " .. node.col)
    end
    return t, tenv
end

function checktype.call(node, tenv)
    local targs = checktype(node.args, tenv)
    local tfunc = especialize_func(checktype(node.func, tenv),
                                   targs, node.lin, node.col)
    if tfunc.tag == 'tfunc' then 
        if #tfunc.params.seq == #targs then
            for pos, tparam in ipairs(tfunc.params.seq) do
                if not tequal(tparam, 
                              especialize(targs[pos], tparam, node.lin, node.col)) then
                   error("o tipo do parametro " .. ptype(tparam) ..
                    " não bate com tipo do argumento " .. 
                    ptype(targs[pos]) ..
                    " sendo o argumento " .. pos .. " na linha " ..
                    node.lin .." coluna " .. node.col)
                end        
            end
            return tfunc.ret, tenv
        else 
             error("o numero de parametros não bate, era esperado " ..
                #tfunc.params.seq .. ", mas foi passado " ..#targs .. 
                " na linha " .. node.lin .." coluna " .. node.col)
        end
    elseif tfunc.tag == "tprim" then
        return checkprim[tfunc.prim](targs, node.lin, node.col), tenv
    else
        error("não é uma função, é " .. ptype(tfunc)..
            " na linha " .. node.lin
              .." coluna " .. node.col)
    end    
end

function checktype.exps(node, tenv)
    local targs = {}
    for _, arg in ipairs(node.seq) do
        targs[#targs + 1] =  checktype(arg, tenv)
    end
    return targs
end

function checktype.set(node, tenv)
    local trval = checktype(node.rval, tenv)
    local tlval = tenv:lookup(node.lval, node.lin, node.col)
    if tequal(trval, tlval) then
       return ast.tunit{ lin = node.lin, col = node.col },
              tenv
    else
        error("tipo do lado direito " .. ptype(trval) .. 
            " não bate com tipo do lado esquerda " .. ptype(tlval) ..
            " na linha " .. node.lin .. " coluna " .. node.col)
    end
end

function checktype.num(node, tenv)
    return ast.tnum{ lin = node.lin, col = node.col }
end

function checktype.str(node, tenv)
    return ast.tstr{ lin = node.lin, col = node.col }
end

function checktype.id(node, tenv)
    return tenv:lookup(node.name, node.lin, node.col)
end

checktype["true"] = function (node, tenv)
    return ast.tbool{ lin = node.lin, col = node.col }
end

checktype["false"] = function (node, tenv)
    return ast.tbool{ lin = node.lin, col = node.col }
end

local function checktype_arith(node, tenv)
    local tleft = checktype(node.left, tenv)
    if not tequal(tleft, "tnum") then
        error("tipo do operando esquerdo da operação aritm. " ..
            ptype(tleft) .. " não é número na linha " .. node.lin ..
            " coluna " .. node.col)
    end
    local tright = checktype(node.right, tenv)
    if not tequal(tright, "tnum") then
        error("tipo do operando direito da operação aritm. " ..
            ptype(tright) .. " não é número na linha " .. node.lin ..
            " coluna " .. node.col)
    end
    return ast.tnum{ lin = node.lin, col = node.col }
end

local function checktype_rel(node, tenv)
    local tleft = checktype(node.left, tenv)
    if not tequal(tleft, "tnum") then
        error("tipo do operando esquerdo da operação rel. " ..
            tleft .. " não é número na linha " .. node.lin ..
            " coluna " .. node.col)
    end
    local tright = checktype(node.right, tenv)
    if not tequal(tright, "tnum") then
        error("tipo do operando direito da operação rel. " ..
            tleft .. " não é número na linha " .. node.lin ..
            " coluna " .. node.col)
    end
    return ast.tbool{ lin = node.lin, col = node.col }
end

local function checktype_log(node, tenv)
    local tleft = checktype(node.left, tenv)
    if not tequal(tleft, "tbool") then
        error("tipo do operando esquerdo da operação lógica " ..
            tleft .. " não é booleano na linha " .. node.lin ..
            " coluna " .. node.col)
    end
    local tright = checktype(node.right, tenv)
    if not tequal(tright, "tbool") then
        error("tipo do operando direito da operação lógica " ..
            tleft .. " não é booleano na linha " .. node.lin ..
            " coluna " .. node.col)
    end
    return ast.tbool{ lin = node.lin, col = node.col }
end

checktype.add = checktype_arith
checktype.sub = checktype_arith
checktype.mul = checktype_arith
checktype.div = checktype_arith

checktype.lt = checktype_rel
checktype.eq = checktype_rel
checktype.neq = checktype_rel

checktype["or"] = checktype_log
checktype["and"] = checktype_log

checktype["not"] = function (node, tenv)
end

function checktype.concat(node, tenv)
    local tleft = checktype(node.left, tenv)
    local tright = checktype(node.right, tenv)
    if tleft.tag == "tgen" and tright.tag ~= "tgen" then
        tleft = especialize(tleft, tright, node.lin, node.col)
    end
    if tright.tag == "tgen" and tleft.tag ~= "tgen" then
        tright = especialize(tright, tleft, node.lin, node.col)
    end
    if tleft.tag == "tseq" or tright.tag == "tseq" then
        if not tequal(tleft, tright) then
            error("tipos não batem na concatenação, lado esquerdo " ..
                ptype(tleft) .. " e lado direito " .. 
                ptype(tright)  ..
                " na linha " .. node.lin .. " coluna " .. node.col)
        end
    else
        if not tequal(tleft, "tstr") then
            error("tipo do operando esquerdo da concatenação " ..
                tleft .. " não é string na linha " .. node.lin ..
                " coluna " .. node.col)
        end
        if not tequal(tright, "tstr") then
            error("tipo do operando direito da concatenação " ..
                tleft .. " não é string na linha " .. node.lin ..
                " coluna " .. node.col)
        end
    end
    return tleft, tenv
end

return checktype
